# -*- coding: utf-8 -*-
"""
Created on Wed Aug 30 19:43:42 2017

@author: Alvaro

Modified on Jun 2021 by Jose Manuel Martin Cabeza

"""
import numpy as np
from math import sqrt, sin, cos, pi, atan
from matplotlib.axes._axes import Axes as mpAxes
import matplotlib.pyplot as plt
import matplotlib.patches as patches
import scipy.linalg as sp

from StruPy2D.config import TOL
from StruPy2D.Node import NodeSol
from StruPy2D.Bar import BarSol

from StruPy2D.drawBC import drawBC, drawElasSup, drawDispSup
from StruPy2D.drawLoad import drawLoad
import StruPy2D.Graphics.Axes as spAxes
from StruPy2D.Graphics.graphNode import graphNode as gN
from StruPy2D.Graphics.graphBar import graphBar as gB
from StruPy2D.Graphics.graphLoad import graphLoad as gL

from StruPy2D.config import TOL
from StruPy2D.Load import Load, LoadGroup, LoadList
from StruPy2D.Node import NodeList
from StruPy2D.Bar import BarList
from StruPy2D.Material import MatList, defMat
from StruPy2D.Section import SecList, defSec
from StruPy2D.Graphics.Style import defNodeStyle, defBarStyle, defLoadStyle

class Static:
    def __init__(self,S,LG=[],run=False):
        # Comprobación de que S es una estructura
        if(type(S).__name__ != "Structure"):
            raise TypeError("First argument of Static constructor must be a Structure object")
            
        self.S = S
        
        self.LG = []
        if(type(LG).__name__ == "LoadGroup"):
            self.LG.append(LG)
            
        elif(type(LG).__name__ == 'list'):
            for lg in LG:
                if(type(lg).__name__ == "LoadGroup"):
                    self.LG.append(lg)
                else:
                    raise TypeError("List of LoadGroups must contain only LoadGroup objects")
        else:
            raise TypeError("LG must be a LoadGroup or list of LoadGroups")
            
        self.ok = False
        self.UD = [] # Solución en desplazamientos
        self.U = [] # Todo el vector de desplazamientos de la estructura
        self.R = [] # Solución en reacciones
        self.F = [] # Vector de todas las fuerzas nodales 
        
        # Soluciones a nivel local, para desplz., reacciones, esfuerzos y tensiones
        self.nds = [] # Lista de objetos solución nodo a nodo
        self.bars = [] # Lista de objetos solución barra a barra
        
        if(run):
            self.run()
            self.ok = True
            
    def addLG(self,LG):
        if(type(LG).__name__ == "LoadGroup"):
            self.LG.append(LG)
            
        elif(type(LG).__name__ == 'list'):
            for lg in LG:
                if(type(lg).__name__ == "LoadGroup"):
                    self.LG.append(lg)
                else:
                    raise TypeError("List of LoadGroups must contain only LoadGroup objects")
        else:
            raise TypeError("LG must be a LoadGroup or list of LoadGroups")
            
    def delLG(self,LG):
        if(type(LG).__name__ == 'LoadGroup' and LG in self.LG):
            self.LG.remove(LG)
        else:
            print("LoadGroup not found. No LoadGroup was deleted.")
            
    def run(self):
        """Ejecuta el cálculo estático de la estructura S."""
        
        self.S.getMatrices()
        self.S.getFeq()
        
        # Aplicación de las condiciones de contorno
        BC = self.S.BC
        nBC = [not x for x in BC]
        
        KDD = self.S.Kg[np.ix_(BC, BC)]
        KCC = self.S.Kg[np.ix_(nBC, nBC)]
        KDC = self.S.Kg[np.ix_(BC, nBC)]
        KCD = self.S.Kg[np.ix_(nBC, BC)]
        
        FD = self.S.Feq[BC] + self.S.F[BC]
        FC = self.S.Feq[nBC] + self.S.F[nBC]
        
        # Solución del sistema de ecuaciones
        self.UD = np.linalg.solve(KDD, FD - (KDC @ self.S.UC))
        self.R = KCC @ self.S.UC + KCD @ self.UD - FC
        self.UD[np.abs(self.UD) < TOL*0.01] = 0
        self.R[np.abs(self.R) < TOL] = 0
        
        # Reconstrucción de los vectores de desplazamientos y fuerzas nodales
        self.U = np.zeros((3*len(self.S.nds),1))
        self.U[BC] = self.UD
        self.U[nBC] = self.S.UC
        
        self.F = np.zeros((3*len(self.S.nds),1))
        self.F[nBC] = self.R
        self.F += self.S.F
        
        self.generateSolution()
        
        self.ok = True
        
    def generateSolution(self):    
        """Genera el conjunto de objetos NodeSol y BarSol que contienen la
        solución completa de la estructura."""
        self.nds = [] # Lista de objetos solución nodo a nodo
        self.bars = [] # Lista de objetos solución barra a barra
        
        for i in range(len(self.S.nds)):
            if(self.S.nds[i].th != 0):
                N = self.S.nds[i]
                th = N.th
                L = np.mat([[cos(th),sin(th),0],[-sin(th),cos(th),0],[0,0,1]])
                
                ind = [3*i, 3*i+1, 3*i+2]
                self.U[ind] = L.T @ self.U[ind]
                self.F[ind] = L.T @ self.F[ind]
                
            self.nds.append(NodeSol(self.S.nds[i],self.U[3*i:3*i+3],self.F[3*i:3*i+3]))
            
        
        for i in range(len(self.S.bars)):
            in1 = self.S.nds.index(self.S.bars[i].p1)
            in2 = self.S.nds.index(self.S.bars[i].p2)
            
            UG = np.append(self.nds[in1].ug,self.nds[in2].ug,0)
            
            self.bars.append(BarSol(self.S.bars[i],self.nds[in1],self.nds[in2],UG))
                  
        
        
    def show(self, axes = None, bc_scale = 0.04, k_scale = 0.1, f_scale = 0.075,
             showNodes = False, node_style = defNodeStyle, bar_style = defBarStyle,
             load_style = defLoadStyle, escaleDef=30):
        # bc_scale : Escala de las condiciones de contorno   
        
        """Crea un pequeño GUI para estudiar la solución obtenida."""
        """Visualización de la estructura"""
        if(axes is None):
            ax = spAxes.SPAxes()
        elif(isinstance(axes, spAxes.SPAxes)):
            ax = axes
        elif(isinstance(axes, mpAxes)):
            pos = axes.get_position()
            ax = spAxes.SPAxes(axes.figure, [pos.x0,pos.y0,pos.width,pos.height])
        else:
            raise TypeError('Axes must be a matplotlib.pyplot.Axes or StruPy2D.Axes.SPAxes object.')
        
        """Cálculo del tamaño característico de la estructura."""
        x_list = self.S.nds.x()
        y_list = self.S.nds.y()
        tam = sqrt((min(y_list)-max(y_list))**2 + (min(x_list)-max(x_list))**2)
        
        """Aplicación de este factor a algunas dimensiones de los estilos"""
        load_style.dim = f_scale*tam if (load_style.dim is None) else load_style.dim
        
        """Visualización de nodos"""
        for N in self.S.nds:
           #ax.plot(N.x,N.y,'bs',markerfacecolor = node_style.MarkerFaceColor ,ms = node_style.MarkerSize, zorder = 1)
           node_style.visible = showNodes
           ax.addElement(gN(N, node_style))
        
        """Visualización de barras con sus libertades"""
        for B in self.S.bars:
            ax.addElement(gB(B, bar_style))
            
        for N in self.nds:
            Lbc = bc_scale * tam
            Lk = k_scale * tam
            Lf = f_scale * tam
            drawBC(ax,N.node,Lbc)
            
            if(True in [x!=0 for x in N.node.k]):
                drawElasSup(ax,N.node,Lk)             
            if(True in [x != 0 for x in N.node.u]):
                drawDispSup(ax,N.node,Lbc)
            #if(len(N.node.loads) > 0):
                #drawLoad(ax,N,Lf)
                #ax.addElement(gL(N.node.loads[0], load_style, ax))   
                  
            
        """Deformada de la estructura en el modo elegido"""
           
        self.deformateStatic(escaleDef)
        
     
        plt.axis('equal')
        plt.title(r'Deformada de la estructura a estático')
                                          
        if(axes == None):
            return ax
        else:
            return None    
            
    
        
        
    def deformateStatic(self,escale):
        
        """ Funcion para la visualización de la deformada del calculo a estatico"""
        for B in self.bars:  
            

            #matriz de cambio de base para x,y
            Lred = np.matrix(np.array([[ cos(B.bar.th), sin(B.bar.th)],
                                    [-sin(B.bar.th), cos(B.bar.th)],]))
            
            #correccion de longitud de cada barra con factor de escala
            Ldelta = B.bar.L-escale*float(B.ul[0])+escale*float(B.ul[3])  
            
            
            #solucion en coord locales de la deformada de la barra         
            xx = np.linspace(0,Ldelta,21)
            
            aa = np.matrix([[0**3, 0**2, 0, 1],
                  [Ldelta**3, Ldelta**2, Ldelta, 1],
                  [3*0**2, 2*0, 1, 0],
                  [3*Ldelta**2, 2*Ldelta, 1, 0]])
           
            bb = np.matrix(np.array([[escale*B.ul[1,0]],
                                     [escale*B.ul[4,0]],
                                     [escale*B.ul[2,0]],
                                     [escale*B.ul[5,0]]])).reshape((len(aa),1))
              
            co = sp.solve(aa,bb)
            yy = (co[0]*xx**3 + co[1]*xx**2 + co[2]*xx + co[3])
 
                
            #giro y desplazamiento de la solucion en coord locales al modelo global
            XX = []
            YY = []
            
            for j in range (len(xx)):
                
                a=float(B.ug[0])
                b=float(B.ug[1])

                xy = np.array([xx[j],yy[j]-yy[0]])
                xy = np.linalg.inv(Lred)*xy.reshape(2,1)

                XX.append(float(xy[0])+B.bar.p1.x+escale*a)
                YY.append((float(xy[1])+B.bar.p1.y+escale*b))
            
           
                #Representacion de la solucion en el grafico global
            
            plt.plot(XX,YY,'r',lw=2)

            
      
        
    